package SliMP3::Background;

# SliMP3 Server Copyright (C) 2001 Sean Adams, Slim Devices Inc.
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License, 
# version 2.

#---------------------------------------
#
# Background tasks - see also Scheduler.pm
#
# maybe there should be a directory for background tasks, and a separate 
# .pm file for each one.
#
#---------------------------------------

use strict;

use Class::Struct;
use File::Spec::Functions qw(:ALL);
use SliMP3::Misc;


#--------------------------------------------------------------------

#
# Example: spins in the background for 10 seconds
# You can see this working if you watch the server with "top"
#
my %test_start_time = ();

sub test_start {
	my $client = shift;

	$test_start_time{$client} = Time::HiRes::time();

	SliMP3::Scheduler::add_task(\&test_run, $client);
}

sub test_kill {
	#todo
}

sub test_run {
	my $client = shift;

	# return 0 when done
        if (Time::HiRes::time() < $test_start_time{$client} + 10) {
		return 1;
	} else {
		&test_done;
		return 0;
	}
};

sub test_done {
}	

#--------------------------------------------------------------------------

#
# Scans the given directory tree (treating .m3u files as directories).
#

my $ds = 0;  # debugging on/off

struct( scandir_dirState => [
	path	 => '$',  # path to this dir
	contents => '@',  # list of the items in this directory
	numcontents => '$', 
	index    => '$',  # our position in contents[]
]);

struct( scandir_clientState => [
	stack	=> '@',		# each level in the hierarchy
	numstack => '$',	# number of levels on the stack
	items	=> '@',		# the big list of all the items we've scanned
	numitems => '$',	# number of items in the list
]);

my %scandirClients = ();

sub scandir_start {
	my ($client, $basepath) = @_;

	# Initialize the stack to one level, the base directory, with
	# index==-1 to indicate we haven't read the dir yet.

	my $basedir = scandir_dirState -> new();
	$basedir->path($basepath);
	$basedir->numcontents(0);
	$basedir->index(-1);

	$scandirClients{$client} = scandir_clientState -> new();
	$scandirClients{$client}->stack(0, $basedir);
	$scandirClients{$client}->numstack(1);
	$scandirClients{$client}->numitems(0);

	SliMP3::Scheduler::add_task(\&scandir_run, $client);
}	

# todo put this somehwere else
#
sub fileNameOK {
	my $filename = shift;

	my $bad = (	($filename eq '') ||
			($filename =~/^\./)
	);

	$bad && $ds && msg("bad: $filename\n");

	return !$bad;
}
		
sub scandir_run {
	my $client = shift;

	my $clientState = $scandirClients{$client};
	my $stackRef = $clientState->stack;
	my $curdirState = @$stackRef[$clientState->numstack-1];	# The directory we're currently in is stack[numstack-1]

	$ds && msg("numitems: ".$clientState->numitems."\n");

######### index==-1 means we need to read the directory

	if ($curdirState->index == -1) {
		my $diritem;
		my $i=0;

		$ds && msg("opendir: ".$curdirState->path."\n");
		opendir DIR, $curdirState->path;
		foreach $diritem (sort(readdir(DIR))) {
			$curdirState->contents($i++, $diritem);
		}
		$curdirState->index(0);
		$curdirState->numcontents($i);
		return 1;
	}

######### OK, the directory has been opened, and index points to the entry we should look at
 
	my $item = '';
	my $ok = 0;

	# move index to the next "good" item
	do {
		if ($curdirState->index == $curdirState->numcontents) {
			#### We've made it to the end of this directory. Pop it from the stack.
			$ds && msg("END OF DIR\n");
			$clientState->numstack($clientState->numstack - 1);

			#### If the stack is empty, we're done!
			if ($clientState->numstack) {
				return 1;
			} else {
				&scandir_done($client, $clientState->items);
				return 0;
			}
		}

		$item = $curdirState->contents($curdirState->index);
		$curdirState->index($curdirState->index + 1);		

	} until (fileNameOK($item));


	my $itempath = catdir($curdirState->path, $item);
	$ds && msg("itempath: $itempath\n");

######### if it's a directory, push it onto the stack

	if (SliMP3::Info::isDir($itempath)) {
		my $newdir = scandir_dirState -> new();
		$newdir->path($itempath);
		$newdir->index(-1);

		$clientState->stack($clientState->numstack, $newdir);
		$clientState->numstack($clientState->numstack + 1);
		return 1;
	}

######### it's a file; add it to the list

	$ds && msg("SCANDIR ADDING: $itempath\n");
	$clientState->items($clientState->numitems, $itempath);
	$clientState->numitems($clientState->numitems + 1);
	return 1;	
}

sub scandir_kill {
}

sub scandir_done {
	my ($client, $items_ref) = @_;

	$ds && msg("Here's what I got:\n", join("\n", @$items_ref));
}	

1;
