package SliMP3::Display;

# $Id: Display.pm,v 1.60 2003/02/11 18:48:24 blackketter Exp $

# SliMP3 Server Copyright (C) 2001 Sean Adams, Slim Devices Inc.
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License, 
# version 2.

use strict;
use File::Spec::Functions qw(:ALL);
use SliMP3::Misc;
use SliMP3::VFD;
use SliMP3::Timers;
use SliMP3::Playlist;
use SliMP3::Strings qw(string);
use POSIX qw(strftime);

# Kill the animation and display what we expect to be showing
sub update {
	my $client = shift;
	
	SliMP3::Animation::killAnimation($client);
	SliMP3::VFD::vfdUpdate($client, curLines($client));
}

sub renderOverlay {
	my $line1 = shift;
	my $line2 = shift;
	my $overlay1 = shift;
	my $overlay2 = shift;
	
	if (defined($overlay1)) {
		my $overlayLength =  length($overlay1);
		$line1 .= ' ' x 40;
		$line1 = substr($line1, 0, 40 - $overlayLength) . $overlay1;
	}
	
	if (defined($overlay2)) {
		my $overlayLength =  length($overlay2);
		$line2 .= ' ' x 40;
		$line2 = substr($line2, 0, 40 - $overlayLength) . $overlay2;
	}
	return ($line1, $line2);
}

# the lines functions return a pair of lines and a pair of overlay strings
# which may need to be overlayed on top of the first pair, right justified.

sub curLines {
	my $client = shift;
	
	my $linefunc = $client->lines();
	
	if (defined $linefunc) {
		return renderOverlay(&$linefunc($client));
	} else {
		$::d_ui && msg("Linefunction for client is undefined!\n");
		$::d_ui && bt();
	}
}

#	FUNCTION:	volumeDisplay
#
#	DESCRIPTION:	Used to display a bar graph of the current volume below a label
#	
#	EXAMPLE OUTPUT:	Volume
#					###############-----------------
#	
#	USAGE:		volumeDisplay($client)
sub volumeDisplay {
	my $client = shift;
	SliMP3::Animation::showBriefly($client,SliMP3::Buttons::Settings::volumeLines($client));
}

sub center {
	my $line = shift;

	return (" " x int((40 - length $line)/2)) . $line;
}

# the following are the custom character definitions for the new progress/level bar...

my @leftchar = ( [ 0b00000111, 
				   0b00001000, 
				   0b00010000, 
				   0b00010000, 
				   0b00010000, 
				   0b00001000, 
				   0b00000111, 
				   0b00000000 ],
				 [ 0b00000111, 
				   0b00001000, 
				   0b00011000, 
				   0b00011000, 
				   0b00011000, 
				   0b00001000, 
				   0b00000111, 
				   0b00000000 ],
				 [ 0b00000111, 
				   0b00001100, 
				   0b00011100, 
				   0b00011100, 
				   0b00011100, 
				   0b00001100, 
				   0b00000111, 
				   0b00000000 ],
				 [ 0b00000111, 
				   0b00001110, 
				   0b00011110, 
				   0b00011110, 
				   0b00011110, 
				   0b00001110, 
				   0b00000111, 
				   0b00000000 ],
				 [ 0b00000111, 
				   0b00001111, 
				   0b00011111, 
				   0b00011111, 
				   0b00011111, 
				   0b00001111, 
				   0b00000111, 
				   0b00000000 ]);
my @middlechar =([ 0b00011111, 
				   0b00000000, 
				   0b00000000, 
				   0b00000000, 
				   0b00000000, 
				   0b00000000, 
				   0b00011111, 
				   0b00000000 ],
				 [ 0b00011111, 
				   0b00010000, 
				   0b00010000, 
				   0b00010000, 
				   0b00010000, 
				   0b00010000, 
				   0b00011111, 
				   0b00000000 ],
				 [ 0b00011111, 
				   0b00011000, 
				   0b00011000, 
				   0b00011000, 
				   0b00011000, 
				   0b00011000, 
				   0b00011111, 
				   0b00000000 ],
				 [ 0b00011111, 
				   0b00011100, 
				   0b00011100, 
				   0b00011100, 
				   0b00011100, 
				   0b00011100, 
				   0b00011111, 
				   0b00000000 ],
				 [ 0b00011111, 
				   0b00011110, 
				   0b00011110, 
				   0b00011110, 
				   0b00011110, 
				   0b00011110, 
				   0b00011111, 
				   0b00000000 ],
				 [ 0b00011111, 
				   0b00011111, 
				   0b00011111, 
				   0b00011111, 
				   0b00011111, 
				   0b00011111, 
				   0b00011111, 
				   0b00000000 ]);

my @rightchar = ([ 0b00011100, 
				   0b00000010, 
				   0b00000001, 
				   0b00000001, 
				   0b00000001, 
				   0b00000010, 
				   0b00011100, 
				   0b00000000 ],
				 [ 0b00011100, 
				   0b00010010, 
				   0b00010001, 
				   0b00010001, 
				   0b00010001, 
				   0b00010010, 
				   0b00011100, 
				   0b00000000 ],
				 [ 0b00011100, 
				   0b00011010, 
				   0b00011001, 
				   0b00011001, 
				   0b00011001, 
				   0b00011010, 
				   0b00011100, 
				   0b00000000 ],
				 [ 0b00011100, 
				   0b00011110, 
				   0b00011101, 
				   0b00011101, 
				   0b00011101, 
				   0b00011110, 
				   0b00011100, 
				   0b00000000 ],
				 [ 0b00011100, 
				   0b00011110, 
				   0b00011111, 
				   0b00011111, 
				   0b00011111, 
				   0b00011110, 
				   0b00011100, 
				   0b00000000 ]);

# returns progress bar text AND sets up custom characters if necessary
sub progressBar {
	my $client = shift;
	my $width = shift;
	my $fractioncomplete = shift;
	
	if ($fractioncomplete < 0) {
		$fractioncomplete = 0;
	}
	
	if ($fractioncomplete > 1.0) {
		$fractioncomplete = 1.0;
	}
	
	if ($width == 0) {
		return "";
	}
	
	my $chart = "";
	
	my $totaldots = scalar @leftchar + ($width - 2) * (scalar @middlechar) + scalar @rightchar;

	my $dots = int($fractioncomplete * $totaldots);
	
	if ($dots < 0) { $dots = 0 };
	
	if ($dots < scalar @leftchar) {
		SliMP3::VFD::setCustomChar($client, 0, @{$leftchar[$dots]});
	} else {
		SliMP3::VFD::setCustomChar($client, 0, @{$leftchar[-1]});
	}
	
	$dots -= scalar @leftchar;
	$chart = $SliMP3::VFD::CUSTOM0;
			
	for (my $i = 1; $i < ($width - 1); $i++) {
		if ($dots <= 0) {
			SliMP3::VFD::setCustomChar ($client, 2, @{$middlechar[0]});
			$chart .= $SliMP3::VFD::CUSTOM2;					
		} elsif ($dots < scalar @middlechar) {
			SliMP3::VFD::setCustomChar ($client, 3, @{$middlechar[$dots]});
			$chart .= $SliMP3::VFD::CUSTOM3;								
		} else {
			SliMP3::VFD::setCustomChar ($client, 4, @{$middlechar[-1]});
			$chart .= $SliMP3::VFD::CUSTOM4;								
		}
		$dots -= scalar @middlechar;
	}
	
	if ($dots <= 0) {
		SliMP3::VFD::setCustomChar($client, 5, @{$rightchar[0]});
	} elsif ($dots < scalar @rightchar) {
		SliMP3::VFD::setCustomChar($client, 5, @{$rightchar[$dots]});
	} else {
		SliMP3::VFD::setCustomChar($client, 5, @{$rightchar[-1]});
	}
	$chart .= $SliMP3::VFD::CUSTOM5;								
	
	return $chart;
}

my @leftmark = ( 		0b00011111,
						0b00000001,
						0b00000001,
						0b00000001,
						0b00000001,
						0b00000001,
						0b00011111,
						0b00000000   );
                  
my @rightmark = ( 		0b00011111,
						0b00010000,
						0b00010000,
						0b00010000,
						0b00010000,
						0b00010000,
						0b00011111,
						0b00000000   );
                  
# returns a +/- balance/bass/treble bar text AND sets up custom characters if necessary
# range 0 to 100, 50 is middle.
sub balanceBar {
	my $client = shift;
	my $width = shift;
	my $balance = shift;
	
	if ($balance < 0) {
		$balance = 0;
	}
	
	if ($balance > 100) {
		$balance = 100;
	}
	
	if ($width == 0) {
		return "";
	}
	
	my $chart = "";
		
	my $edgepos = $balance / 100.0 * $width;
	# do the leftmost char
	if ($balance <= 0) {
		SliMP3::VFD::setCustomChar($client, 0, @{$leftchar[-1]});
	} else {
		SliMP3::VFD::setCustomChar($client, 0, @{$leftchar[0]});	
	}
	$chart = $SliMP3::VFD::CUSTOM0;
	
	SliMP3::VFD::setCustomChar($client, 1, @{$middlechar[0]});
	SliMP3::VFD::setCustomChar($client, 2, @{$middlechar[-1]});
	SliMP3::VFD::setCustomChar($client, 6, @leftmark);
	SliMP3::VFD::setCustomChar($client, 7, @rightmark);

	my $i;
	
	# left half
	for ($i = 1; $i < $width/2; $i++) {
		if ($balance >= 50) {
			if ($i == $width/2-1) {
				$chart .= $SliMP3::VFD::CUSTOM6;
			} else {
				$chart .= $SliMP3::VFD::CUSTOM1;
			}
		} else {
			if ($i < $edgepos) {
				$chart .= $SliMP3::VFD::CUSTOM1;
			} else {
				$chart .= $SliMP3::VFD::CUSTOM2;
			}			
		}
	}
	
	# right half
	for ($i = $width/2; $i < $width-1; $i++) {
		if ($balance <= 50) {
			if ($i == $width/2) {
				$chart .= $SliMP3::VFD::CUSTOM7;
			} else {	
				$chart .= $SliMP3::VFD::CUSTOM1;
			}
		} else {
			if ($i < $edgepos) {
				$chart .= $SliMP3::VFD::CUSTOM2;
			} else {
				$chart .= $SliMP3::VFD::CUSTOM1;
			}			
		}
	}
	
	# do the rightmost char
	if ($balance >= 100) {
		SliMP3::VFD::setCustomChar($client, 3, @{$rightchar[-1]});
	} else {
		SliMP3::VFD::setCustomChar($client, 3, @{$rightchar[0]});	
	}
	$chart .= $SliMP3::VFD::CUSTOM3;
	
	return $chart;
}

# replaces ~ in format string
my @toplinechar = (		0b00011111, 
						0b00000000, 
						0b00000000, 
						0b00000000, 
						0b00000000, 
						0b00000000, 
						0b00000000, 
						0b00000000	 );

# replaces = in format string
my @doublelinechar = (	0b00011111, 
						0b00000000, 
						0b00000000, 
						0b00000000, 
						0b00000000, 
						0b00000000, 
						0b00011111, 
						0b00000000	 );

# replaces ? in format string.  Used in Z, ?, 7
my @Ztop = (      		0b00011111,
						0b00000001,
						0b00000001,
						0b00000010,
						0b00000100,
						0b00001000,
						0b00010000,
						0b00000000   );
                  
# replaces < in format string.  Used in Z, 2, 6
my @Zbottom = (   		0b00000001,
						0b00000010,
						0b00000100,
						0b00001000,
						0b00010000,
						0b00010000,
						0b00011111,
						0b00000000   );
                  
# replaces / in format string.
my @slash = (     		0b00000001,
						0b00000001,
						0b00000010,
						0b00000100,
						0b00001000,
						0b00010000,
						0b00010000,
						0b00000000   );
                  
my @backslash = ( 		0b00010000,
						0b00010000,
						0b00001000,
						0b00000100,
						0b00000010,
						0b00000001,
						0b00000001,
						0b00000000   );
                  
my @filledcircle = ( 	0b00000001,
						0b00001111,
						0b00011111,
						0b00011111,
						0b00011111,
						0b00001110,
						0b00000000,
						0b00000000   );
                  
my $leftvbar = chr(0x10);
my $rightvbar = chr(0x18);

# special characters that are not handled by the format string
my %specialchars = (
	
	"(" => [ $SliMP3::VFD::CUSTOM5,
			 $SliMP3::VFD::CUSTOM6 ],
	
	")" => [ ' ' . $SliMP3::VFD::CUSTOM6,
			 ' ' . $SliMP3::VFD::CUSTOM5 ],
	
	"[" => [ $rightvbar . $SliMP3::VFD::CUSTOM1,
			 $rightvbar . '_' ],
	
	"]" => [ $SliMP3::VFD::CUSTOM1 . $leftvbar,
			 '_' . $leftvbar],
	
	"<" => [ '/',
			 '\\' ],
	
	">" => [ '\\',
			 '/' ],
	
	"{" => [ '(',
			 '(' ],
	
	"}" => [ ')',
			 ')' ],
	
	'"' => [ '\'\'',
			 '  '],
	"%" => [ 'o/', '/o'],
	"&" => [ '_' . 'L', $SliMP3::VFD::CUSTOM6 . $leftvbar],
	"^" => [ $SliMP3::VFD::CUSTOM5 . $SliMP3::VFD::CUSTOM6, '  '],
	" " => [ '  ', '  ' ],
	"'" => [ '|', ' ' ],
	"!" => [ '|', '.' ],
	":" => [ '.', '.' ],
	"." => [ ' ', '.' ],
	";" => [ '.', ',' ],
	"," => [ ' ', '/' ],
	"`" => [ '\\', ' ' ],
	
	"_" => [ '  ', '_' . '_' ],
	
	"+" => [ '_' . 'L', ' ' . $leftvbar],
	
	"*" => [ '**', '**'],
	
	'~' => [ $SliMP3::VFD::CUSTOM5 . $SliMP3::VFD::CUSTOM1, '  ' ],
	
	"@" => [ $SliMP3::VFD::CUSTOM5 . 'd',
			 $SliMP3::VFD::CUSTOM6 . '_' ],
	
	"#" => [ '##', '##' ],
	
	'$' => [ '$$', '$$' ],
	
	"|" => [ '|',
			 '|' ],
	
	"-" => [ '_' . '_',
			 '  ' ],
	
	"/" => [ ' ' . $SliMP3::VFD::CUSTOM5,
			 $SliMP3::VFD::CUSTOM5 . ' ' ],
	
	"\\" => [ $SliMP3::VFD::CUSTOM6 . ' ',
			  ' ' . $SliMP3::VFD::CUSTOM6 ],
	
	"=" => ['--'
		   ,'--'],
	
	'?' => [$SliMP3::VFD::CUSTOM1 . $SliMP3::VFD::CUSTOM4,
		,' .'],
		
	chr(0x10) => [ chr(0x10), chr(0x10) ],
	chr(0x11) => [ chr(0x11), chr(0x11) ],
	chr(0x12) => [ chr(0x12), chr(0x12) ],
	chr(0x13) => [ chr(0x13), chr(0x13) ],
	chr(0x14) => [ chr(0x14), chr(0x14) ],
	chr(0x15) => [ chr(0x15), chr(0x15) ],
	chr(0x16) => [ chr(0x16), chr(0x16) ],
	chr(0x17) => [ chr(0x17), chr(0x17) ],
	chr(0x18) => [ chr(0x18), chr(0x18) ],
	
	chr($SliMP3::VFD::NOTESYMBOL) => [ chr(0x18) . $SliMP3::VFD::CUSTOM6 , $SliMP3::VFD::CUSTOM2 . " "],

	chr($SliMP3::VFD::RIGHTARROW) => [ ' _\\' , ' ' . $SliMP3::VFD::CUSTOM1 . '/']
);

# the font format string
my $double = 

	# all digits are 3 chars wide
	'0/~\01 /[12 ~)23 =)34]_[45]=~56 < 67 ~?78(=)89(=)9' .
	'0\_/01  [12 <_23 _)34  [45 _)56(_)67 / 78(_)89 / 9' .
	
 	# kerning is custom so exclude blanks here except for 'I'
 	'A /\ AB]=)BC/~CD]~\DE]=EF]=FG/~ GH]_[HI [IJ  [J' .
 	'A]~~[AB]_)BC\_CD]_/DE]_EF] FG\=[GH] [HI [IJ]_[J' .
	
	'K]/KL] LM]\/[MN]\[NO/~\OP]=)PQ/~\QR]=)RS(~S' .
	'K]\KL]_LM]  [MN] [NO\_/OP]  PQ\_xQR] \RS_)S' .
	
	'T~|~TU] [UV[]VW[  ]WX\/XY\/YZ~?Z' .
	'T | TU]_[UV\/VW\/\/WX/\XY [YZ<_Z';
	
my $kernL = '\~\]\?\_\<\=';
my $kernR = '\~\[\<\_\\\\/';

#
# double the height and width of line 2 of the display
#
sub doubleSize {
	my $client = shift;
	my ($line1, $line2) = (shift,shift);
	my ($newline1, $newline2) = ("", "");
	
	if (!defined($line2) || $line2 eq "") { $line2 = $line1; };
	
	$line2 =~ s/^(\s*)(.*)/$2/;
	
	# setup the special characters
	SliMP3::VFD::setCustomChar($client, 0, @Zbottom);
	SliMP3::VFD::setCustomChar($client, 1, @toplinechar);
	SliMP3::VFD::setCustomChar($client, 2, @filledcircle);
	SliMP3::VFD::setCustomChar($client, 3, @doublelinechar);
	SliMP3::VFD::setCustomChar($client, 4, @Ztop);
	SliMP3::VFD::setCustomChar($client, 5, @slash);
	SliMP3::VFD::setCustomChar($client, 6, @backslash);
	
	$::d_ui && msg("undoubled line1: $line1\n");
	$::d_ui && msg("undoubled line2: $line2\n");
	$line2 = uc($line2);
	$line2 =~ s/(?:|)/AE/g;
	$line2 =~ s/(?:|)/OE/g;
	$line2 =~ tr{ݟ}
			{AAAAAACEEEEIIIIDNOOOOOUUUUYYAAAAAACEEEEIIIIDNOOOOOUUUUYY};
	my $lastch1 = "";
	my $lastch2 = "";
   
	my $lastchar = "";
	my $specialcharkeys = join( '', keys %specialchars);
	foreach my $char (split //,$line2) {
		if (index($specialcharkeys, $char) != -1) {
			$newline1 .= @{$specialchars{$char}}[0];
			$newline2 .= @{$specialchars{$char}}[1];
		} elsif (index($double, $char) != -1) {

			my $pat = "$char(.*)$char.*$char(.*)$char";
			my ($char1, $char2) = $double =~ /$pat/;
			
			if ($char =~ /[A-Z]/) {
				if ($lastchar ne ' ') {
					if (($lastch1 =~ /[$kernL]$/ && $char1 =~ /^[$kernR]/) ||
						 ($lastch2 =~ /[$kernL]$/ && $char2 =~ /^[$kernR]/)) {
					
						if ($lastchar =~ /[CGLSTZ]/ && $char =~ /[COQ]/) {
							# Special cases to exclude kerning between
						} else {
						   $newline1 .= ' ';
						   $newline2 .= ' ';
						}
					}
				}
			}
			
			$lastch1 = $char1;
			$lastch2 = $char2;
			
			$char1 =~ s/~/$SliMP3::VFD::CUSTOM1/g;
			$char1 =~ s/=/$SliMP3::VFD::CUSTOM3/g;
			$char1 =~ s/\?/$SliMP3::VFD::CUSTOM4/g;
			$char1 =~ s/\[/$leftvbar/g;
			$char1 =~ s/\]/$rightvbar/g;
			$char1 =~ s/\//$SliMP3::VFD::CUSTOM5/g;
			$char1 =~ s/\\/$SliMP3::VFD::CUSTOM6/g;
			$char1 =~ s/</$SliMP3::VFD::CUSTOM0/g;
			$char1 =~ s/x/X/g;  # use capitol "X" in letter "Q"
			
			$char2 =~ s/~/$SliMP3::VFD::CUSTOM1/g;
			$char2 =~ s/=/$SliMP3::VFD::CUSTOM3/g;
			$char2 =~ s/\?/$SliMP3::VFD::CUSTOM4/g;
			$char2 =~ s/\[/$leftvbar/g;
			$char2 =~ s/\]/$rightvbar/g;
			$char2 =~ s/\//$SliMP3::VFD::CUSTOM5/g;
			$char2 =~ s/\\/$SliMP3::VFD::CUSTOM6/g;
			$char2 =~ s/</$SliMP3::VFD::CUSTOM0/g;
			$char2 =~ s/x/X/g;  # use capitol "X" in letter "Q"
			
			$newline1 .= $char1;
			$newline2 .= $char2;
			
		} else {
			# warn "unknown character '$char' in double mode";
			next;
		}
		$lastchar = $char;
	}

	$newline1 = $newline1 . (' ' x (40 - length($newline1)));
	$newline2 = $newline2 . (' ' x (40 - length($newline1)));

	return ($newline1, $newline2);

}

1;
__END__


